/* sds_io.c: SDS 940 I/O simulator

   Copyright (c) 2001, Robert M. Supnik

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   ROBERT M SUPNIK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of Robert M Supnik shall not
   be used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from Robert M Supnik.
*/

#include "sds_defs.h"

#define TST_XFR(d)	(xfr_req && (1 << dev_map[d]))
#define SET_XFR(d)	xfr_req = xfr_req | (1 << dev_map[d])
#define CLR_XFR(d)	xfr_req = xfr_req & ~(1 << dev_map[d])
#define INV_DEV(d)	(dev_map[d] == 0)
#define VLD_DEV(d)	(dev_map[d] != 0)
#define TST_EOR(c)	(chan[c].flag & CHF_EOR)
#define QAILCE(a)	(((a) >= POT_ILCY) && ((a) < (POT_ILCY + NUM_CHAN)))
struct channel chan[NUM_CHAN];				/* channels */
static const char *chname[NUM_CHAN + 1] = { "ANY",
	"W", "Y", "C", "D", "E", "F", "G", "H" };
extern uint32 M[MAXMEMSIZE];				/* memory */
extern uint32 int_req;					/* int req */
extern uint32 xfr_req;					/* xfer req */
extern uint32 alert;					/* pin/pot alert */
extern uint32 X, EM2, EM3, OV, ion, bpt;
extern uint32 nml_mode, usr_mode, rtc_pie;
extern int32 stop_invins, stop_invdev, stop_inviop;
extern int32 mon_usr_trap;
extern UNIT cpu_unit;

t_stat chan_reset (UNIT *uptr);
t_stat chan_read (int32 ch);
t_stat chan_write (int32 ch);
void chan_write_mem (int32 ch);
void chan_flush_war (int32 ch);
uint32 chan_mar_inc (int32 ch);
t_stat chan_eor (int32 ch);
t_stat pot_ilc (uint32 num, uint32 *dat);
t_stat pot_dcr (uint32 num, uint32 *dat);
t_stat pin_adr (uint32 num, uint32 *dat);
extern t_stat pot_RL1 (uint32 num, uint32 *dat);
extern t_stat pot_RL2 (uint32 num, uint32 *dat);
extern t_stat pot_RL4 (uint32 num, uint32 *dat);
extern t_stat pin_fhds (uint32 num, uint32 *dat);
extern t_stat pot_fhda (uint32 num, uint32 *dat);
extern t_stat pin_rad (uint32 num, uint32 *dat);
extern t_stat pot_rad (uint32 num, uint32 *dat);
extern t_stat tti_conn (uint32 inst, int32 mode);
extern t_stat tti_disc (uint32 dev);
extern t_stat tti_read (uint32 dev, uint32 *dat);
extern t_stat tto_conn (uint32 inst, int32 mode);
extern t_stat tto_disc (uint32 dev);
extern t_stat tto_write (uint32 dev, uint32 *dat);
extern t_stat ptr_conn (uint32 inst, int32 mode);
extern t_stat ptr_disc (uint32 dev);
extern t_stat ptr_read (uint32 dev, uint32 *dat);
extern t_stat ptp_conn (uint32 inst, int32 mode);
extern t_stat ptp_disc (uint32 dev);
extern t_stat ptp_write (uint32 dev, uint32 *dat);
extern t_stat lpt_conn (uint32 inst, int32 mode);
extern t_stat lpt_disc (uint32 dev);
extern t_stat lpt_write (uint32 dev, uint32 *dat);
extern t_stat lpt_wreor (uint32 dev);
extern t_bool lpt_sks (uint32 inst);
extern t_stat fhd_conn (uint32 inst, int32 mode);
extern t_stat fhd_disc (uint32 dev);
extern t_stat fhd_read (uint32 dev, uint32 *dat);
extern t_stat fhd_write (uint32 dev, uint32 *dat);
extern t_stat fhd_wreor (uint32 dev);
extern t_bool fhd_sks (uint32 inst);
extern t_stat rad_conn (uint32 inst, int32 mode);
extern t_stat rad_disc (uint32 dev);
extern t_stat rad_read (uint32 dev, uint32 *dat);
extern t_stat rad_write (uint32 dev, uint32 *dat);
extern t_stat rad_wreor (uint32 dev);
extern t_bool rad_sks (uint32 inst);
extern t_stat mt_conn (uint32 inst, int32 mode);
extern t_stat mt_disc (uint32 dev);
extern t_stat mt_read (uint32 dev, uint32 *dat);
extern t_stat mt_write (uint32 dev, uint32 *dat);
extern t_stat mt_wreor (uint32 dev);
extern t_bool mt_sks (uint32 inst);
extern void set_dyn_map (void);

/* SDS I/O model

   A device is modeled by its interactions with a channel.  Devices can only be
   accessed via channels.  Each channel has its own device address space.  This
   means devices can only be accessed from a specific channel.  The simulator
   allows greater flexibility.  Devices can be assigned to specific channel; or
   they can be set up to respond on all channels.

   I/O operations start with a channel connect.  The EOM instruction is passed
   to the device via the conn routine.  This routine is also used for non-channel
   EOM's to the device.  For channel connects, the device must remember the
   channel number.

   The device responds (after a delay) by setting its XFR_RDY flag.  This causes
   the channel to invoke either the read or write routine (for input or output)
   to get or put the next character.  If the device is an asynchronous output
   device, it calls routine chan_set_ordy to see if there is output available.
   If there is, XFR_RDY is set; if not, the channel is marked to wake the
   attached device when output is available.  This prevents invalid rate errors.

   Output may be terminated by a write end of record, a disconnect, or both.
   Write end of record occurs when the word count reaches zero on an IORD or IORP
   operation.  It also occurs if a TOP instruction is issued.  The device is
   expected to respond by setting the end of record indicator in the channel,
   which will in turn trigger an end of record interrupt.

   When the channel operation completes, the channel disconnects and calls the
   disconnect processor to perform any device specific cleanup.  The differences
   between write end of record and disconnect are subtle.  On magtape output,
   for example, both signal end of record; but write end of record allows the
   magtape to continue moving, while disconnect halts its motion.

   Finally, there are specific routines to handle SKS, and, for devices that
   support it, PIN and POT.

   Any valid device (i.e., a device with a non-zero entry in dev_map) must supply
   connect, disconnect, and read or write routines.  SKS, PIN, POT, and write end
   of record routines are optional.

   Channels could, optionally, handle 12b or 24b characters.  The simulator can
   support all modes.
*/
   
struct iodisp {
	t_stat	(*conn) ();				/* inst, mode */
	t_stat	(*disc) ();				/* dev */
	t_stat	(*read) ();				/* dev, *dat */
	t_stat	(*write) ();				/* dev, *dat */
	t_stat	(*wreor) ();				/* dev */
	t_bool	(*sks) ();				/* dev */
	};

struct aldisp {
	t_stat	(*pin) ();				/* altnum, *dat */
	t_stat	(*pot) ();				/* altnum, *dat */
	};

/* Channel data structures

   chan_dev	channel device descriptor
   chan_unit	channel unit descriptor
   chan_reg	channel register list
*/

UNIT chan_unit = { UDATA (NULL, 0, 0) };

REG chan_reg[] = {
	{ ORDATA (W_UAR, chan[CHAN_W].uar, 6) },
	{ ORDATA (W_WCR, chan[CHAN_W].wcr, 15) },
	{ ORDATA (W_MAR, chan[CHAN_W].mar, 16) },
	{ ORDATA (W_DCR, chan[CHAN_W].dcr, 6) },
	{ ORDATA (W_WAR, chan[CHAN_W].war, 24) },
	{ ORDATA (W_CPW, chan[CHAN_W].cpw, 2) },
	{ ORDATA (W_CNT, chan[CHAN_W].cnt, 3) },
	{ ORDATA (W_MODE, chan[CHAN_W].mode, 12) },
	{ ORDATA (W_FLAG, chan[CHAN_W].flag, CHF_N_FLG) },
	{ ORDATA (Y_UAR, chan[CHAN_Y].uar, 6) },
	{ ORDATA (Y_WCR, chan[CHAN_Y].wcr, 15) },
	{ ORDATA (Y_MAR, chan[CHAN_Y].mar, 16) },
	{ ORDATA (Y_DCR, chan[CHAN_Y].dcr, 6) },
	{ ORDATA (Y_WAR, chan[CHAN_Y].war, 24) },
	{ ORDATA (Y_CPW, chan[CHAN_Y].cpw, 2) },
	{ ORDATA (Y_CNT, chan[CHAN_Y].cnt, 3) },
	{ ORDATA (Y_MODE, chan[CHAN_Y].mode, 12) },
	{ ORDATA (Y_FLAG, chan[CHAN_Y].flag, CHF_N_FLG) },
	{ ORDATA (C_UAR, chan[CHAN_C].uar, 6) },
	{ ORDATA (C_WCR, chan[CHAN_C].wcr, 15) },
	{ ORDATA (C_MAR, chan[CHAN_C].mar, 16) },
	{ ORDATA (C_DCR, chan[CHAN_C].dcr, 6) },
	{ ORDATA (C_WAR, chan[CHAN_C].war, 24) },
	{ ORDATA (C_CPW, chan[CHAN_C].cpw, 2) },
	{ ORDATA (C_CNT, chan[CHAN_C].cnt, 3) },
	{ ORDATA (C_MODE, chan[CHAN_C].mode, 12) },
	{ ORDATA (C_FLAG, chan[CHAN_C].flag, CHF_N_FLG) },
	{ ORDATA (D_UAR, chan[CHAN_D].uar, 6) },
	{ ORDATA (D_WCR, chan[CHAN_D].wcr, 15) },
	{ ORDATA (D_MAR, chan[CHAN_D].mar, 16) },
	{ ORDATA (D_DCR, chan[CHAN_D].dcr, 6) },
	{ ORDATA (D_WAR, chan[CHAN_D].war, 24) },
	{ ORDATA (D_CPW, chan[CHAN_D].cpw, 2) },
	{ ORDATA (D_CNT, chan[CHAN_D].cnt, 3) },
	{ ORDATA (D_MODE, chan[CHAN_D].mode, 12) },
	{ ORDATA (D_FLAG, chan[CHAN_D].flag, CHF_N_FLG) },
	{ ORDATA (E_UAR, chan[CHAN_E].uar, 6) },
	{ ORDATA (E_WCR, chan[CHAN_E].wcr, 15) },
	{ ORDATA (E_MAR, chan[CHAN_E].mar, 16) },
	{ ORDATA (E_DCR, chan[CHAN_E].dcr, 6) },
	{ ORDATA (E_WAR, chan[CHAN_E].war, 24) },
	{ ORDATA (E_CPW, chan[CHAN_E].cpw, 2) },
	{ ORDATA (E_CNT, chan[CHAN_E].cnt, 3) },
	{ ORDATA (E_MODE, chan[CHAN_E].mode, 12) },
	{ ORDATA (E_FLAG, chan[CHAN_E].flag, CHF_N_FLG) },
	{ ORDATA (F_UAR, chan[CHAN_F].uar, 6) },
	{ ORDATA (F_WCR, chan[CHAN_F].wcr, 15) },
	{ ORDATA (F_MAR, chan[CHAN_F].mar, 16) },
	{ ORDATA (F_DCR, chan[CHAN_F].dcr, 6) },
	{ ORDATA (F_WAR, chan[CHAN_F].war, 24) },
	{ ORDATA (F_CPW, chan[CHAN_F].cpw, 2) },
	{ ORDATA (F_CNT, chan[CHAN_F].cnt, 3) },
	{ ORDATA (F_MODE, chan[CHAN_F].mode, 12) },
	{ ORDATA (F_FLAG, chan[CHAN_F].flag, CHF_N_FLG) },
	{ ORDATA (G_UAR, chan[CHAN_G].uar, 6) },
	{ ORDATA (G_WCR, chan[CHAN_G].wcr, 15) },
	{ ORDATA (G_MAR, chan[CHAN_G].mar, 16) },
	{ ORDATA (G_DCR, chan[CHAN_G].dcr, 6) },
	{ ORDATA (G_WAR, chan[CHAN_G].war, 24) },
	{ ORDATA (G_CPW, chan[CHAN_G].cpw, 2) },
	{ ORDATA (G_CNT, chan[CHAN_G].cnt, 3) },
	{ ORDATA (G_MODE, chan[CHAN_G].mode, 12) },
	{ ORDATA (G_FLAG, chan[CHAN_G].flag, CHF_N_FLG) },
	{ ORDATA (H_UAR, chan[CHAN_H].uar, 6) },
	{ ORDATA (H_WCR, chan[CHAN_H].wcr, 15) },
	{ ORDATA (H_MAR, chan[CHAN_H].mar, 16) },
	{ ORDATA (H_DCR, chan[CHAN_H].dcr, 6) },
	{ ORDATA (H_WAR, chan[CHAN_H].war, 24) },
	{ ORDATA (H_CPW, chan[CHAN_H].cpw, 2) },
	{ ORDATA (H_CNT, chan[CHAN_H].cnt, 3) },
	{ ORDATA (H_MODE, chan[CHAN_H].mode, 12) },
	{ ORDATA (H_FLAG, chan[CHAN_H].flag, CHF_N_FLG) },
	{ NULL }  };

DEVICE chan_dev = {
	"CHAN", &chan_unit, chan_reg, NULL,
	1, 8, 8, 1, 8, 8,
	NULL, NULL, NULL,
	NULL, NULL, &chan_reset };

/* Tables */

static const uint32 int_zc[8] = {
	INT_WZWC, INT_YZWC, INT_CZWC, INT_DZWC,
	INT_EZWC, INT_FZWC, INT_GZWC, INT_HZWC };

static const uint32 int_er[8] = {
	INT_WEOR, INT_YEOR, INT_CEOR, INT_DEOR,
	INT_EEOR, INT_FEOR, INT_GEOR, INT_HEOR };

static const int32 dev_map[64] = {
	0, XFR_V_TTI, 0, 0, XFR_V_PTR, 0, 0, 0,
	XFR_V_MT0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, XFR_V_FHD, XFR_V_RAD,
	XFR_V_MT0, 0, 0, 0, 0, 0, 0, 0,
	0, XFR_V_TTO, 0, 0, XFR_V_PTP, 0, 0, 0,
	XFR_V_MT0, 0, 0, 0, 0, 0, 0, 0,
	XFR_V_LPT, 0, 0, 0, 0, 0, XFR_V_FHD, XFR_V_RAD,
	XFR_V_MT0, 0, 0, 0, 0, 0, 0, 0  };

struct iodisp dev_dsp[64] = {
	{ NULL, NULL, NULL, NULL, NULL, NULL },		/* 00 - 17 */
	{ &tti_conn, &tti_disc, &tti_read, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ &ptr_conn, &ptr_disc, &ptr_read, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ NULL, NULL, NULL, NULL, NULL, NULL },		/* 20 - 37 */
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ &fhd_conn, &fhd_disc, &fhd_read, NULL, &fhd_wreor, &fhd_sks },
	{ &rad_conn, &rad_disc, &rad_read, NULL, &rad_wreor, &rad_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, &mt_read, NULL, &mt_wreor, &mt_sks },
	{ NULL, NULL, NULL, NULL, NULL, NULL },		/* 40 - 57 */
	{ &tto_conn, &tto_disc, NULL, &tto_write, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ &ptp_conn, &ptp_disc, NULL, &ptp_write, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
							/* 60 - 77 */
	{ &lpt_conn, &lpt_disc, NULL, &lpt_write, &lpt_wreor, &lpt_sks },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ NULL, NULL, NULL, NULL, NULL, NULL },
	{ &fhd_conn, &fhd_disc, NULL, &fhd_write, &fhd_wreor, &fhd_sks },
	{ &rad_conn, &rad_disc, NULL, &rad_write, &rad_wreor, &rad_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks },
	{ &mt_conn, &mt_disc, NULL, &mt_write, &mt_wreor, &mt_sks }  };

struct aldisp dev_alt[] = {
	{ NULL, NULL },
	{ NULL, &pot_ilc }, { NULL, &pot_ilc },
	{ NULL, &pot_ilc }, { NULL, &pot_ilc }, 
	{ NULL, &pot_ilc }, { NULL, &pot_ilc },
	{ NULL, &pot_ilc }, { NULL, &pot_ilc }, 
	{ NULL, &pot_dcr }, { NULL, &pot_dcr },
	{ NULL, &pot_dcr }, { NULL, &pot_dcr }, 
	{ NULL, &pot_dcr }, { NULL, &pot_dcr },
	{ NULL, &pot_dcr }, { NULL, &pot_dcr },
	{ &pin_adr, NULL }, { &pin_adr, NULL },
	{ &pin_adr, NULL }, { &pin_adr, NULL },
	{ &pin_adr, NULL }, { &pin_adr, NULL },
	{ &pin_adr, NULL }, { &pin_adr, NULL },
	{ NULL, &pot_RL1 }, { NULL, &pot_RL2 },
	{ NULL, &pot_RL4 },
	{ &pin_fhds, NULL }, { NULL, &pot_fhda },
	{ &pin_rad, &pot_rad }  };

/* Single word I/O instructions */

t_stat op_wyim (uint32 inst, uint32 *dat)
{
int32 ch, dev;

ch = (inst & 000200000)? CHAN_W: CHAN_Y;		/* get chan# */
dev = chan[ch].uar & DEV_MASK;				/* get dev # */
if (dev == 0) CRETIOP;					/* none? err */
if (chan[ch].cnt <= chan[ch].cpw) return STOP_IONRDY;	/* buf empty? hang */
*dat = chan[ch].war;					/* get data */
chan[ch].cnt = 0;					/* buffer empty */
return SCPE_OK;
}

t_stat op_miwy (uint32 inst, uint32 dat)
{
int32 ch, dev;

ch = (inst & 000200000)? CHAN_W: CHAN_Y;		/* get chan# */
dev = chan[ch].uar & DEV_MASK;				/* get dev # */
if (dev == 0) CRETIOP;					/* none? err */
if (chan[ch].cnt) return STOP_IONRDY;			/* buf full? hang */
chan[ch].war = dat;					/* store data */
chan[ch].cnt = chan[ch].cpw + 1;			/* buffer full */
if (chan[ch].flag & CHF_OWAK) {				/* output wake? */
	if (VLD_DEV (dev)) SET_XFR (dev);		/* wake channel */
	chan[ch].flag = chan[ch].flag & ~CHF_OWAK;  }	/* clear wake */
return SCPE_OK;
}

t_stat op_pin (uint32 *dat)
{
uint32 al = alert;					/* local copy */

alert = 0;						/* clear alert */
if ((al == 0) || (dev_alt[al].pin == NULL)) CRETIOP;	/* inv alert? */
return dev_alt[al].pin (al, &dat);			/* PIN from dev */
}

t_stat op_pot (uint32 dat)
{
uint32 al = alert;					/* local copy */

alert = 0;						/* clear alert */
if ((al == 0) || (dev_alt[al].pin == NULL)) CRETIOP;	/* inv alert? */
return dev_alt[al].pot (al, &dat);			/* POT to dev */
}

/* EOM/EOD */

t_stat op_eomd (uint32 inst)
{
uint32 mod = I_GETIOMD (inst);				/* get mode */
uint32 ch = I_GETEOCH (inst);				/* get chan # */
uint32 dev = inst & DEV_MASK;				/* get dev # */
uint32 ch_dev = chan[ch].uar & DEV_MASK;		/* chan curr dev # */
t_stat r;

switch (mod) {
case 0:							/* IO control */
	if (dev) {					/* new dev? */
		if (ch_dev) CRETIOP;			/* chan act? err */
		if (INV_DEV (dev)) CRETDEV;		/* inv dev? err */
		if (((dev & DEV_OUT)?			/* inv op? err */
		     dev_dsp[dev].write: dev_dsp[dev].read) == NULL)
			CRETIOP;
		chan[ch].war = chan[ch].cnt = 0;	/* init chan */
		chan[ch].flag = chan[ch].dcr = 0;
		chan[ch].uar = 0;
		chan[ch].mode = (ch > CHAN_E)? CHM_CE: 0;
		if (r = dev_dsp[dev].conn (inst, 0)) return r;	/* connect */
		if ((inst & I_IND) || (ch >= CHAN_C)) {	/* C-H? alert ilc */
			alert = POT_ILCY + ch;
			chan[ch].mar = chan[ch].wcr = 0;  }
		if (chan[ch].flag & CHF_24B)		/* 24B? 1 ch/wd */
			 chan[ch].cpw = 0;
		else if (chan[ch].flag & CHF_12B)	/* 12B? 2 ch/wd */
			chan[ch].cpw = CHC_GETCPW (inst & 1);
		else chan[ch].cpw = CHC_GETCPW (inst);	/* 6b, 1-4 ch/wd */
		chan[ch].uar = dev;			/* connected */
		if ((dev & DEV_OUT) && ion && !QAILCE (alert))
			int_req = int_req | int_zc[ch];  }
	else if (VLD_DEV (ch_dev))			/* disconnect */
		return dev_dsp[ch_dev].disc (ch_dev);
	break;
case 1:							/* buf control */
	if (QAILCE (alert)) {				/* ilce alerted? */
		uint32 prvch = alert - POT_ILCY;	/* get prv chan */
		if (prvch >= CHAN_E) inst = inst | CHM_CE; /* DACC? ext */
		chan[ch].mode = inst;			/* save mode */
		chan[ch].mar = (CHM_GETHMA (inst) << 14) |
			(chan[ch].mar & CHI_M_MA);	/* get hi mar */
		chan[ch].wcr = (CHM_GETHWC (inst) << 10) |
			(chan[ch].wcr & CHI_M_WC);  }	/* get hi wc */
	else if (dev) {					/* dev EOM */
		if (INV_DEV (dev)) CRETDEV;		/* inv dev? err */
		if (r = dev_dsp[dev].conn (inst, 1)) return r;  }
	else {	inst = inst & 047677;			/* chan EOM */
		if (inst == 040000) {			/* alert ilace */
			alert = POT_ILCY + ch;
			chan[ch].mar = chan[ch].wcr = 0;  }
		else if (inst == 001000)		/* alert DCR */
			alert = POT_DCRY + ch;
		else if (inst == 002000)		/* alert addr */
			alert = POT_ADRY + ch;
		else if (inst == 004000) {		/* term output */
			if (dev_dsp[dev].wreor) {	/* send eor */		
				r = dev_dsp[dev].wreor (ch_dev);
				if (r) return r;  }  }  }
	break;		 
case 2:							/* internal */
	if (ch >= CHAN_E) {				/* EOD? */
		if (inst & 00300) {			/* set EM? */
			if (inst & 00100) EM2 = inst & 07;
			if (inst & 00200) EM3 = (inst >> 3) & 07;
			set_dyn_map ();
			break;  }
		else CRETINS;  }			/* end if EOD */
	if (inst & 00001) OV = 0;			/* clr OV */
	if (inst & 00002) ion = 1;			/* ion */
	else if (inst & 00004) ion = 0;			/* iof */
	if ((inst & 00010) && (((X >> 1) ^ X) & EXPS)) OV = 1;
	if (inst & 00100) rtc_pie = 1;			/* arm clk pls */
	else if (inst & 00200) rtc_pie = 0;		/* disarm pls */
	if ((inst & 01400) == 01400) alert = POT_RL4;	/* alert RL4 */
	else if (inst & 00400) alert = POT_RL1;		/* alert RL1 */
	else if (inst & 01000) alert = POT_RL2;		/* alert RL2 */
	if (inst & 02000) {				/* nml to mon */
		nml_mode = usr_mode = 0;
		if (inst & 00400) mon_usr_trap = 1;  }
	break;
default:						/* inv instr */
	CRETINS;  }					/* end case */
return SCPE_OK;
}

/* Skip if not signal */

t_stat op_sks (uint32 inst, uint32 *dat)
{
uint32 mod = I_GETIOMD (inst);				/* get mode */
uint32 ch = I_GETSKCH (inst);				/* get chan # */
uint32 dev = inst & DEV_MASK;				/* get dev # */

*dat = 0;
if ((ch == 2) && !(inst & 037774)) {			/* EM test */
	if (((inst & 0001) && (EM2 != 2)) ||
	    ((inst & 0002) && (EM2 != 3))) *dat = 1;
	return SCPE_OK;  }
switch (mod) {
case 1:							/* ch, dev */
	if (dev) {					/* device */
		if (INV_DEV (dev)) CRETDEV;		/* inv dev? err */
		if (dev_dsp[dev].sks)			/* sense status */
			*dat = dev_dsp[dev].sks (dev);  }
	else {						/* channel */
		if (((inst & 04000) && (chan[ch].uar == 0)) ||
		    ((inst & 02000) && (chan[ch].wcr == 0)) ||
		    ((inst & 01000) && ((chan[ch].flag & CHF_ERR) == 0)) ||
		    ((inst & 00400) && (chan[ch].flag & CHF_IREC))) *dat = 1;  }
	break;
case 2:							/* internal test */
	if ((inst & 0001) && OV) {			/* test OV */
		*dat = 1;
		OV = 0;					/* and reset */
		break;  }
	if (((inst & 0002) && !ion) ||			/* ion, bpt test */
	    ((inst & 0004) && ion) ||
	    ((inst & 0040) && ((bpt & 001) == 0)) ||
	    ((inst & 0100) && ((bpt & 002) == 0)) ||
	    ((inst & 0200) && ((bpt & 004) == 0)) ||
	    ((inst & 0400) && ((bpt & 010) == 0))) *dat = 1;
	break;
default:						/* inv instr */
	CRETINS;  }
return SCPE_OK;
}

/* PIN/POT routines */

t_stat pot_ilc (uint32 num, uint32 *dat)
{
uint32 ch = num - POT_ILCY;

chan[ch].mar = (chan[ch].mar & ~CHI_M_MA) | CHI_GETMA (*dat);
chan[ch].wcr = (chan[ch].wcr & ~CHI_M_WC) | CHI_GETWC (*dat);
return SCPE_OK;
}

t_stat pot_dcr (uint32 num, uint32 *dat)
{
uint32 ch = num - POT_DCRY;

chan[ch].dcr = (*dat) & (CHD_INT | CHD_PAGE);
chan[ch].flag = chan[ch].flag | CHF_DCHN;
return SCPE_OK;
}

t_stat pin_adr (uint32 num, uint32 *dat)
{
uint32 ch = num - POT_ADRY;

*dat = chan[ch].mar & PAMASK;
return SCPE_OK;
}

/* Channel read invokes the I/O device to get the next character and,
   if not end of record, assembles it into the word assembly register.
   If the interlace is on, the full word is stored in memory.
   The key difference points for the various terminal functions are

	end of record	comp: EOT interrupt
			IORD, IOSD: EOR interrupt, disconnect
			IORP, IOSP: EOR interrupt, interrecord
	interlace off:	comp: EOW interrupt
			IORD, IORP: ignore
			IOSD, IOSP: overrun error
	--wcr == 0:	comp: clear interlace
			IORD, IORP, IOSP: ZWC interrupt
			IOSD: ZWC interrupt, EOR interrupt, disconnect

   Note that the channel can be disconnected if CHN_EOR is set, but must
   not be if XFR_REQ is set.
*/
			
t_stat chan_read (int32 ch)
{
uint32 dat = 0;
uint32 dev = chan[ch].uar & DEV_MASK;
uint32 tfnc = CHM_GETFNC (chan[ch].mode);
t_stat r = SCPE_OK;

if (dev && TST_XFR (dev)) {				/* ready to xfr? */
    if (dev_dsp[dev].read == NULL) CRETIOP;		/* can't read? */
    r = dev_dsp[dev].read (dev, &dat);			/* read data */
    if (r) chan[ch].flag = chan[ch].flag | CHF_ERR;	/* error? */
    if (chan[ch].flag & CHF_24B) chan[ch].war = dat;	/* 24B? */
    else if (chan[ch].flag & CHF_12B)			/* 12B? */
	chan[ch].war = ((chan[ch].war << 12) | (dat & 07777)) & DMASK;
    else chan[ch].war = ((chan[ch].war << 6) | (dat & 077)) & DMASK;
    chan[ch].cnt = chan[ch].cnt + 1;			/* insert char */
    if (chan[ch].cnt > chan[ch].cpw) {			/* full? */
	if (chan[ch].flag & CHF_SCAN)			/* scanning? */
	    chan[ch].cnt = chan[ch].cpw;		/* never full */
	else if (chan[ch].flag & CHF_ILCE) {		/* interlace on? */
	    chan_write_mem (ch);			/* write to mem */
	    if (chan[ch].wcr == 0) {			/* wc zero? */
		chan[ch].flag = chan[ch].flag & ~CHF_ILCE; /* clr interlace */
		if ((tfnc != CHM_COMP) && (chan[ch].mode & CHM_ZC))
		    int_req = int_req | int_zc[ch];	/* zwc interrupt */
		if (tfnc == CHM_IOSD) {			/* IOSD? also EOR */
		    if (chan[ch].mode & CHM_ER) int_req = int_req | int_er[ch];
		    dev_dsp[dev].disc (dev);		/* disconnect */
		    chan[ch].uar = 0;  }		/* end if IOSD */
		}					/* end if wcr == 0 */
	    }						/* end if ilce on */
	else {						/* interlace off */
	    if (TST_EOR (ch)) return chan_eor (ch);	/* eor? */
	    if ((tfnc == CHM_COMP) && ion)		/* C: EOW intr */
		int_req = int_req | int_zc[ch];
	    else if (tfnc & CHM_SGNL)			/* Sx: error */
		chan[ch].flag = chan[ch].flag | CHF_ERR;
	    else chan[ch].cnt = chan[ch].cpw;		/* Rx: ignore */
	    }						/* end else ilce */
	}						/* end if full */
    }							/* end if xfr */
if (TST_EOR (ch)) {					/* end record? */
    if (tfnc == CHM_COMP) chan_flush_war (ch);		/* C: fill war */
    else if (chan[ch].cnt) {				/* RX, CX: fill? */
	chan_flush_war (ch);				/* fill war */
	if (chan[ch].flag & CHF_ILCE)			/* ilce on? store */
	    chan_write_mem (ch);  }			/* end else if cnt */
    return chan_eor (ch);  }				/* eot/eor int */
return r;		
}

void chan_write_mem (int32 ch)
{
WriteP (chan[ch].mar, chan[ch].war);			/* write to mem */
chan[ch].mar = chan_mar_inc (ch);			/* incr mar */
chan[ch].wcr = (chan[ch].wcr - 1) & 077777;		/* decr wcr */
chan[ch].cnt = 0;					/* reset war */
return;
}

void chan_flush_war (int32 ch)
{
int32 i = (chan[ch].cpw - chan[ch].cnt) + 1;

if (i) {
	if (chan[ch].flag & CHF_24B) chan[ch].war = 0;
	else if (chan[ch].flag & CHF_12B && i)
		chan[ch].war = (chan[ch].war << 12) & DMASK;
	else chan[ch].war = (chan[ch].war << (i * 6)) & DMASK;
	chan[ch].cnt = chan[ch].cpw + 1;  }
return;
}

/* Channel write gets the next character and sends it to the I/O device.
   If this is the last character in an interlace operation, the end of
   record operation is invoked.
   The key difference points for the various terminal functions are

	end of record:	comp: EOT interrupt
			IORD, IOSD: EOR interrupt, disconnect
			IORP, IOSP: EOR interrupt, interrecord
	interlace off:	if not end of record, EOW interrupt
	--wcr == 0:	comp: EOT interrupt, disconnect
			IORD, IORP: ignore
			IOSD: ZWC interrupt, disconnect
			IOSP: ZWC interrupt, interrecord
*/
t_stat chan_write (int32 ch)
{
uint32 dat = 0;
uint32 dev = chan[ch].uar & DEV_MASK;
uint32 tfnc = CHM_GETFNC (chan[ch].mode);
t_stat r = SCPE_OK;

if (dev && TST_XFR (dev)) {				/* ready to xfr? */
    if (dev_dsp[dev].write == NULL) CRETIOP;		/* can't write? */
    if (chan[ch].cnt == 0) {				/* buffer empty? */
	if (chan[ch].flag & CHF_ILCE) {			/* interlace on? */
	    chan[ch].war = ReadP (chan[ch].mar);
	    chan[ch].mar = chan_mar_inc (ch);		/* incr mar */
	    chan[ch].wcr = (chan[ch].wcr - 1) & 077777;	/* decr mar */
	    chan[ch].cnt = chan[ch].cpw + 1;  }		/* set cnt */
	else {						/* ilce off */
	    CLR_XFR (dev);				/* cant xfr */
	    if (TST_EOR (dev)) return chan_eor (ch);	/* EOR? */
	    chan[ch].flag = chan[ch].flag | CHF_ERR;	/* rate err */
	    return SCPE_OK;  }				/* end else ilce */
	}						/* end if cnt */
    chan[ch].cnt = chan[ch].cnt - 1;			/* decr cnt */
    if (chan[ch].flag & CHF_24B) dat = chan[ch].war;	/* 24B? */
    else if (chan[ch].flag & CHF_12B) {			/* 12B? */
	dat = (chan[ch].war >> 12) & 07777;		/* get halfword */
	chan[ch].war = (chan[ch].war << 12) & DMASK;  }	/* remove from war */
    else {						/* 6B */
	dat = (chan[ch].war >> 18) & 077;		/* get char */
	chan[ch].war = (chan[ch].war << 6) & DMASK;  }	/* remove from war */
    r = dev_dsp[dev].write (dev, &dat);			/* write */
    if (r) chan[ch].flag = chan[ch].flag | CHF_ERR;	/* error? */
    if (chan[ch].cnt == 0) {				/* buf empty? */
	if (chan[ch].flag & CHF_ILCE) {			/* ilce on? */
	    if (chan[ch].wcr == 0) {			/* wc now 0? */
		chan[ch].flag = chan[ch].flag & ~CHF_ILCE; /* ilc off */
		if (tfnc == CHM_COMP) {			/* compatible? */
		    if (ion) int_req = int_req | int_zc[ch];
		    dev_dsp[dev].disc (dev);		/* disconnnect */
		    chan[ch].uar = 0;  }		/* end if comp */
		else {					/* extended */
		    if (chan[ch].mode & CHM_ZC)		/* ZWC int */
			int_req = int_req | int_zc[ch];
		    if (tfnc == CHM_IOSD) {		/* SD */
			if (chan[ch].mode & CHM_ER)	/* EOR int */
			    int_req = int_req | int_er[ch];
			dev_dsp[dev].disc (dev);	/* disconnnect */
			chan[ch].uar = 0;  }		/* end if SD */
		    else if (!(tfnc && CHM_SGNL) && dev_dsp[dev].wreor)
			dev_dsp[dev].wreor (dev);	/* R: write EOR */
		    }					/* end else comp */
		}					/* end if wcr */
	    }						/* end if ilce */
	else if (ion) int_req = int_req | int_zc[ch];	/* off, EOW intr */
	}						/* end if cnt */
   }							/* end if xfr */
if (TST_EOR (ch)) return chan_eor (ch);			/* eor rcvd? */
return r;
}

/* MAR increment */

uint32 chan_mar_inc (int32 ch)
{
uint32 t = (chan[ch].mar + 1) & PAMASK;			/* incr mar */
if ((chan[ch].flag & CHF_DCHN) && ((t & VA_POFF) == 0)) {	/* chain? */
	chan[ch].flag = chan[ch].flag & ~CHF_DCHN;	/* clr flag */
	if (chan[ch].dcr & CHD_INT)			/* if armed, intr */
		int_req = int_req | int_zc[ch];
	t = (chan[ch].dcr & CHD_PAGE) << VA_V_PN;  }	/* new mar */
return t;
}

/* End of record action */

t_stat chan_eor (int32 ch)
{
uint32 tfnc = CHM_GETFNC (chan[ch].mode);
uint32 dev = chan[ch].uar & DEV_MASK;

chan[ch].flag = chan[ch].flag & ~(CHF_EOR | CHF_ILCE);	/* clr eor, ilce */
if (((tfnc == CHM_COMP) && ion) || (chan[ch].mode & CHM_ER))
	int_req = int_req | int_er[ch];			/* EOT/EOR? */
if (dev && tfnc & CHM_PROC)				/* P, still conn? */
	chan[ch].flag = chan[ch].flag | CHF_IREC;	/* interrecord */
else {	chan[ch].uar = 0;				/* disconnect */
	if (VLD_DEV (dev)) return dev_dsp[dev].disc (dev);  }
return SCPE_OK;
}

/* Externally visible routines */
/* Channel driver */

t_stat chan_process (void)
{
int32 i;
t_stat r;

for (i = 0; i < NUM_CHAN; i++) {			/* loop thru */
	uint32 dev = chan[i].uar & DEV_MASK;		/* get dev */
	if ((dev && TST_XFR (dev)) || TST_EOR (i)) {	/* chan active? */
		if (dev & DEV_OUT) r = chan_write (i);	/* write */
		else r = chan_read (i);			/* read */
		if (r) return r;  }  }
return SCPE_OK;
}

/* Test for channel active */

t_bool chan_testact (void)
{
int32 i;

for (i = 0; i < NUM_CHAN; i++) {
	int32 dev = chan[i].uar & DEV_MASK;
	if ((dev && TST_XFR (dev)) || TST_EOR (i)) return i;  }
return 0;
}

/* Async output device ready for more data */

void chan_set_ordy (int32 ch)
{
if ((ch >= 0) && (ch < NUM_CHAN)) {
	int32 dev = chan[ch].uar & DEV_MASK;		/* get dev */
	if (chan[ch].cnt || (chan[ch].flag & CHF_ILCE))	/* buf or ilce? */
		SET_XFR (dev);				/* set xfr flg */
	else chan[ch].flag = chan[ch].flag | CHF_OWAK;  } /* need wakeup */
return;
}

/* Set flag in channel */

void chan_set_flag (int32 ch, uint32 fl)
{
if ((ch >= 0) && (ch < NUM_CHAN)) chan[ch].flag = chan[ch].flag | fl;
return;
}

/* Set UAR in channel */

void chan_set_uar (int32 ch, uint32 dev)
{
if ((ch >= 0) && (ch < NUM_CHAN)) chan[ch].uar = dev & DEV_MASK;
return;
}

/* Disconnect channel */

void chan_disc (int32 ch)
{
if ((ch >= 0) && (ch < NUM_CHAN)) chan[ch].uar = 0;
return;
}

/* Reset channels */

t_stat chan_reset (UNIT *uptr)
{
int32 i;

xfr_req = 0;
for (i = 0; i < NUM_CHAN; i++) {
	chan[i].uar = 0;
	chan[i].wcr = 0;
	chan[i].mar = 0;
	chan[i].dcr = 0;
	chan[i].war = 0;
	chan[i].cpw = 0;
	chan[i].cnt = 0;
	chan[i].mode = 0;
	chan[i].flag = 0;  }
return SCPE_OK;
}

/* Channel assignment routines */

t_stat set_chan (UNIT *uptr, int32 val, char *sptr, void *desc)
{
int32 i, *chnp = (int32 *) desc;

if ((chnp == NULL) || (sptr == NULL)) return SCPE_ARG;	/* valid args? */
for (i = 0; i < (NUM_CHAN + 1); i++) {			/* match input */
    if (strcmp (sptr, chname[i]) == 0) {		/* find string */
	if (val && !(val & (1 << i))) return SCPE_ARG;	/* legal? */
	*chnp = i;					/* store new */
	return SCPE_OK;  }  }
return SCPE_ARG;
}

t_stat show_chan (FILE *st, UNIT *uptr, void *desc)
{
int32 *chnp = (int32 *) desc;

if (chnp == NULL) return SCPE_ARG;
if (*chnp > NUM_CHAN) return SCPE_ARG;
fprintf (st, "channel=%s", chname[*chnp]);
return SCPE_OK;
}
